#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""FAL command for RADIUS Extention."""

from iobase.xmldata import (XmlConfNode,)
from iobase.exception import (InvalidArgError, IoBaseException,)

from pal_info import (PAL_RADIUS_PATH, PAL_CONFFILE_PATH,)
import libfal.falcmd as falcmd
import libfal.util as util
from fal_info import (ReqTypes, Results)
import ionas.request as request

from pkgaddon.fal_hradius_config import (CONFIG_FILE,)
from pkgaddon.hradius_config import (RADIUS_SERVICE,)


#Add the service entry for the eset package
util.config.FUNC_TO_SERVICE_MAP[RADIUS_SERVICE] = RADIUS_SERVICE


RADIUS_NAME = "radius"

PERMIT_NAME = "permit"
MACADDR_PATH = PERMIT_NAME + "/macaddrlist/macaddr"
USER_PATH = PERMIT_NAME + "/userlist/user"

AUTHENTICATOR_NAME = "authenticator"
MACADDR_NAME = "macaddr"
USER_NAME = "user"
NAME_NAME = "name"
IPADDR_NAME = "ipaddr"
SECRETKEY_NAME = "secretkey"
PASSWORD_NAME = "password"
HASHED_ATTR = "hashed"

AUTHENTICATE_NAME = "authenticate"
APPLY_TAG_LIST = [AUTHENTICATE_NAME, "permit"]

DELETE_INFO_LIST_MAP = {
        AUTHENTICATE_NAME : [("authenticatorlist", "*")],
        "permit" : [
                    ("macaddrlist", "*"),
                    ("userlist", "*"),
            ],
    }

AUTHENTICATORLIST_PATH = "authenticate/authenticatorlist"
MACADDRLIST_PATH = "permit/macaddrlist"
USERLIST_PATH = "permit/userlist"
AUTHENTICATOR_XPATH_TMPL = \
    AUTHENTICATORLIST_PATH + """/authenticator/name[text()="%s"]"""
MACADDR_XPATH_TMPL = MACADDRLIST_PATH + """/macaddr[text()="%s"]"""
USER_XPATH_TMPL = USERLIST_PATH + """/user/name[text()="%s"]"""

AUTHENTICATOR_XML_TMPL = """<authenticator><name>%s</name>""" + \
            """<ipaddr>%s</ipaddr><secretkey>%s</secretkey></authenticator>"""
MACADDR_XML_TMPL = """<macaddr>%s</macaddr>"""
USER_XML_TMPL = """<user><name>%s</name>""" + \
            """<password hashed="%s">%s</password></user>"""

TRUE_VALUE = "true"
FALSE_VALUE = "false"


class RequestError(IoBaseException):
    """Request error class."""

    pass


class FalRadius(falcmd.FalCmd):
    """FAL RADIUS command implementation class."""

    #
    # Internal methods
    #

    def _check_body_node(self, node, type=None):
        """Check if entry's body node is valid."""

        if node.get_node_name() != RADIUS_NAME:
            raise InvalidArgError(node.get_node_name())

    def _send_request(self, target, entry):
        """Send "entry" request to "target" PAL commands."""

        response_entry = self.send_entry(entry, target)
        if response_entry.get_result() != Results.OK:
            raise RequestError(response_entry.get_result())

        return response_entry.get()

    def _try_deleting_old_setting(self, old_tag):
        """Try deleting old setting in "old_tag"."""

        del_info_list = DELETE_INFO_LIST_MAP.get(old_tag.get_node_name(), [])
        for del_info in del_info_list:
            old_tag.delete_nodes(del_info[0], del_info[1])

    def _is_service_enabled(self, current_config, new_config):
        """Return if authenticate service is enabled.
           Return None if authenticate service state is not changed.
        """

        try:
            current_authenticate = current_config.get_node(AUTHENTICATE_NAME)
            new_authenticate = new_config.get_node(AUTHENTICATE_NAME)

            if (not current_authenticate.is_enabled() and
                    new_authenticate.is_enabled()):

                return True

            if (current_authenticate.is_enabled() and
                    not new_authenticate.is_enabled()):

                return False
        except:
            pass

        return None

    def _add_authenticator_config(self, config, csv_entry):
        """Add authenticator configuration from "csv_entry" to "confg"."""

        if (len(config.get_node_list(AUTHENTICATOR_XPATH_TMPL %
                csv_entry[1].strip())) != 0):
            # A specified authenticator already exists.
            self.warning("already in")
            return

        config.get_node(AUTHENTICATORLIST_PATH).append_child(XmlConfNode(
            AUTHENTICATOR_XML_TMPL % (csv_entry[1].strip(),
                    csv_entry[2].strip(),
                    csv_entry[3].strip())))

    def _add_macaddr_config(self, config, csv_entry):
        """Add macaddr configuration from "csv_entry" to "confg"."""

        if (len(config.get_node_list(MACADDR_XPATH_TMPL %
                csv_entry[1].strip())) != 0):
            # A specified macaddr already exists.
            self.warning("already in")
            return

        config.get_node(MACADDRLIST_PATH).append_child(XmlConfNode(
            MACADDR_XML_TMPL % (csv_entry[1].strip(),)))

    def _add_user_config(self, config, csv_entry):
        """Add user configuration from "csv_entry" to "confg"."""

        if (len(config.get_node_list(USER_XPATH_TMPL %
                csv_entry[1].strip())) != 0):
            # A specified user already exists.
            self.warning("already in")
            return

        is_hashed = FALSE_VALUE
        try:
            if csv_entry[2] != "" and int(csv_entry[2]) > 0:
                is_hashed = TRUE_VALUE
        except:
            pass

        config.get_node(USERLIST_PATH).append_child(XmlConfNode(
            USER_XML_TMPL % (csv_entry[1].strip(),
                    is_hashed,
                    csv_entry[3].strip())))

    def _csvfile_to_config(self, csv_path, config):
        """Update configuration XML for importing "csv_path" file."""

        with open(csv_path) as import_csv:
            for line in import_csv:
                try:
                    csv_entry = line.strip().split(",")
                    if csv_entry[0] == AUTHENTICATOR_NAME:
                        self._add_authenticator_config(config, csv_entry)
                    if csv_entry[0] == MACADDR_NAME:
                        self._add_macaddr_config(config, csv_entry)
                    if csv_entry[0] == USER_NAME:
                        self._add_user_config(config, csv_entry)
                except:
                    self.exception("_csvfile_to_config")
                    # Ignore error
                    continue

        return

    def _get_authenticator_csv_lines(self, config):
        
        csv_lines = ""
        for authenticator in config.get_node_list(
            AUTHENTICATORLIST_PATH + "/" + AUTHENTICATOR_NAME):

            try:
                csv_line = AUTHENTICATOR_NAME + ", "
                csv_line += authenticator.get_value(NAME_NAME) + ", "
                csv_line += authenticator.get_value(IPADDR_NAME) + ", "
                csv_line += authenticator.get_value(SECRETKEY_NAME)
            except:
                self.exception("_get_authenticator_csv_line")
                continue

            csv_lines += csv_line + "\r\n"

        return csv_lines

    def _get_macaddr_csv_lines(self, config):

        csv_lines = ""
        for macaddr in config.get_node_list(
            MACADDRLIST_PATH + "/" + MACADDR_NAME):

            try:
                csv_line = MACADDR_NAME + ", "
                csv_line += macaddr.get_text()
            except:
                self.exception("_get_macaddr_csv_line")
                continue

            csv_lines += csv_line + "\r\n"

        return csv_lines

    def _get_user_csv_lines(self, config):

        csv_lines = ""
        for user in config.get_node_list(
            USERLIST_PATH + "/" + USER_NAME):

            try:
                csv_line = USER_NAME + ", "
                csv_line += user.get_value(NAME_NAME) + ", "
                if user.get_node(PASSWORD_NAME).is_set_to_true(HASHED_ATTR):
                    csv_line += "1, "
                else:
                    csv_line += "0, "
                csv_line += user.get_value(PASSWORD_NAME)
            except:
                self.exception("_get_user_csv_line")
                continue

            csv_lines += csv_line + "\r\n"

        return csv_lines

    def _config_to_csvfile(self, config, csv_path):
        """Return a new configuration XML for importing "csv_path" file."""

        with open(csv_path, "w") as export_csv:
            try:
                export_csv.write(
                    self._get_authenticator_csv_lines(config))
                export_csv.write(
                    self._get_macaddr_csv_lines(config))
                export_csv.write(
                    self._get_user_csv_lines(config))
            except:
                # Ignore error
                pass

        return

    #
    # Request handlers
    #

    def _create_handler(self, entry):
        self._check_body_node(entry.get(), ReqTypes.CREATE)
        try:
            util.servicectrl_suspend([], [RADIUS_SERVICE])
            self._send_request(PAL_RADIUS_PATH, entry);
            util.servicectrl_resume([], [RADIUS_SERVICE])
            entry.set_result(Results.OK)
        except:
            self.exception("create fail")
            entry.set_result(Results.CREATE_ERROR)
        return entry

    def _get_handler(self, entry):
        self._check_body_node(entry.get(), ReqTypes.GET)
        try:
            self._send_request(PAL_RADIUS_PATH, entry);
            entry.set_result(Results.OK)
        except:
            self.exception("get fail")
            entry.set_result(Results.GET_ERROR)
        return entry

    def _delete_handler(self, entry):
        self._check_body_node(entry.get(), ReqTypes.DELETE)
        try:
            util.servicectrl_suspend([], [RADIUS_SERVICE])
            self._send_request(PAL_RADIUS_PATH, entry);
            util.servicectrl_resume([], [RADIUS_SERVICE])
            entry.set_result(Results.OK)
        except:
            self.exception("delete fail")
            entry.set_result(Results.DELETE_ERROR)
        return entry

    def _list_handler(self, entry):
        self._check_body_node(entry.get(), ReqTypes.LIST)
        try:
            resp = self._send_request(PAL_RADIUS_PATH, entry);
            entry.set_result(Results.OK)
            entry.get().append_child(resp)
        except:
            self.exception("list fail")
            entry.set_result(Results.LIST_ERROR)
        return entry

    def _import_handler(self, entry):
        self._check_body_node(entry.get(), ReqTypes.IMPORT)
        try:
            config = util.load_config(CONFIG_FILE)
            self._csvfile_to_config(entry.get_file(), config)
            new_entry = request.Entry(ReqTypes.APPLY, config)
            return self._apply_handler(new_entry)
        except:
            self.exception("import fail")
            entry.set_result(Results.IMPORT_ERROR)
        return entry

    def _export_handler(self, entry):
        self._check_body_node(entry.get(), ReqTypes.EXPORT)
        try:
            config = util.load_config(CONFIG_FILE)
            self._config_to_csvfile(config, entry.get_file())
            entry.set_result(Results.OK)
        except:
            self.exception("export fail")
            entry.set_result(Results.EXPORT_ERROR)
        return entry

    def _apply_handler(self, entry):
        """Handle apply request."""

        self._check_body_node(entry.get(), ReqTypes.APPLY)
        try:
            config = util.load_config(CONFIG_FILE)
            new_config = entry.get()
            
            is_service_enabled = self._is_service_enabled(config, new_config)

            util.servicectrl_suspend([], [RADIUS_SERVICE])
            for tag_name in APPLY_TAG_LIST:
                try:
                    new_target = new_config.get_node(tag_name)
                    try:
                        target = config.get_node(tag_name)
                        self._try_deleting_old_setting(target)
                        target.update(new_target, override=True)
                    except NotFoundError:
                        config.append_child(new_target.get_clone())
                except:
                    # Not change setting about "tag_name" tag entry.
                    continue

            new_entry = request.Entry(ReqTypes.APPLY, config)

            resp = self._send_request(PAL_RADIUS_PATH, new_entry)
            
            for tag_name in APPLY_TAG_LIST:
                try:
                    new_target = resp.get_node(tag_name)
                    try:
                        target = config.get_node(tag_name)
                        self._try_deleting_old_setting(target)
                        target.update(new_target, override=True)
                    except NotFoundError:
                        config.append_child(new_target.get_clone())
                except:
                    # Not change setting about "tag_name" tag entry.
                    continue
                    
            util.servicectrl_resume([], [RADIUS_SERVICE])
            if is_service_enabled is True:
                util.enable_service(RADIUS_SERVICE, enable=True,
                        do_start_stop=True)
            elif is_service_enabled is False:
                util.enable_service(RADIUS_SERVICE, enable=False,
                        do_start_stop=True)

            util.save_config(CONFIG_FILE, config)
            entry.set_result(Results.OK)
        except:
            self.exception("apply fail")
            entry.set_result(Results.APPLY_ERROR)

        return entry


    #
    # Constructor
    #

    def __init__(self, option=None):
        """Constructor."""

        falcmd.FalCmd.__init__(self, option=option)

        load_save_handler = lambda entry: self.send_entry_with_file(
            entry, PAL_CONFFILE_PATH, CONFIG_FILE)

        self.set_handler(ReqTypes.LOAD, load_save_handler)
        self.set_handler(ReqTypes.SAVE, load_save_handler)

        self.set_handler(ReqTypes.APPLY, self._apply_handler)
        self.set_handler(ReqTypes.CREATE, self._create_handler)
        self.set_handler(ReqTypes.GET, self._get_handler)
        self.set_handler(ReqTypes.DELETE, self._delete_handler)
        self.set_handler(ReqTypes.LIST, self._list_handler)
        self.set_handler(ReqTypes.IMPORT, self._import_handler)
        self.set_handler(ReqTypes.EXPORT, self._export_handler)


#
# Main
#

if __name__ == "__main__":
    FalRadius(option=falcmd.FalCmdOption()).handle_request()
