#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""PAL command for RADIUS Extention."""


import os
import shutil

from iobase.exception import (InvalidArgError,)
from iobase.xmldata import (XmlConfNode, XmlNode, )
from iobase.cmdexec import (CmdExec, NULL, PIPE, NotFoundError, ExecError, )
from pal_info import (ReqTypes, Results)
import pal_config as config

import libpal.palcmd as palcmd

from pkgaddon.pal_hradius_config import (RADIUS_CONF_DIR,)

from django.template import Template, Context
from django.conf import settings
from django.utils.text import get_valid_filename
from copy import deepcopy
import collections
import hashlib, binascii
import csv
import re
import glob
from dateutil import parser, tz

if not settings.configured:
    settings.configure()

os.environ["LD_LIBRARY_PATH"] = "/opt/freeradius/lib"

# ========================================================
# Define Constants
# ========================================================
RADIUS_NAME = "radius"
CA_COMMON_NAME = "\"I-O DATA DEVICE, INC.\""
SERVER_COMMON_NAME = "\"landisk\""
OPENSSL_CMD = "/opt/freeradius/bin/openssl"
C_REHASH_CMD = "/opt/freeradius/bin/c_rehash"

# ========================================================
# Templates
# ========================================================
RADIUS_CONF_TEMPLATE_DIR = RADIUS_CONF_DIR + "/templates"
RADIUS_CONF_TEMPLATE_ROOT_RADIUSD_CONF = RADIUS_CONF_TEMPLATE_DIR + "/root.radiusd.conf.tpl"
RADIUS_CONF_TEMPLATE_ROOT_CLIENTS_CONF = RADIUS_CONF_TEMPLATE_DIR + "/root.clients.conf.tpl"
RADIUS_CONF_TEMPLATE_ROOT_AUTHORIZED_MACS = RADIUS_CONF_TEMPLATE_DIR + "/root.authorized_macs.tpl"
RADIUS_CONF_TEMPLATE_ROOT_USERS = RADIUS_CONF_TEMPLATE_DIR + "/root.users.tpl"
RADIUS_CONF_TEMPLATE_MODS_EAP = RADIUS_CONF_TEMPLATE_DIR + "/mods.eap.tpl"
RADIUS_CONF_TEMPLATE_MODS_LOGINTIME = RADIUS_CONF_TEMPLATE_DIR + "/mods.logintime.tpl"
RADIUS_CONF_TEMPLATE_MODS_MSCHAP = RADIUS_CONF_TEMPLATE_DIR + "/mods.mschap.tpl"
RADIUS_CONF_TEMPLATE_MODS_PREPROCESS = RADIUS_CONF_TEMPLATE_DIR + "/mods.preprocess.tpl"
RADIUS_CONF_TEMPLATE_SITES_DEFAULT = RADIUS_CONF_TEMPLATE_DIR + "/sites.default.tpl"

RADIUS_CONF_TEMPLATES = [
    RADIUS_CONF_TEMPLATE_ROOT_RADIUSD_CONF,
    RADIUS_CONF_TEMPLATE_ROOT_CLIENTS_CONF,
    RADIUS_CONF_TEMPLATE_ROOT_AUTHORIZED_MACS,
    RADIUS_CONF_TEMPLATE_ROOT_USERS,
    RADIUS_CONF_TEMPLATE_MODS_EAP,
    RADIUS_CONF_TEMPLATE_MODS_LOGINTIME,
    RADIUS_CONF_TEMPLATE_MODS_MSCHAP,
    RADIUS_CONF_TEMPLATE_MODS_PREPROCESS,
    RADIUS_CONF_TEMPLATE_SITES_DEFAULT,
]

RADIUS_CONF_TEMPLATES_AUTHENTICATE = [
    RADIUS_CONF_TEMPLATE_ROOT_RADIUSD_CONF,
    RADIUS_CONF_TEMPLATE_ROOT_CLIENTS_CONF,
    RADIUS_CONF_TEMPLATE_MODS_EAP,
    RADIUS_CONF_TEMPLATE_MODS_LOGINTIME,
    RADIUS_CONF_TEMPLATE_MODS_MSCHAP,
    RADIUS_CONF_TEMPLATE_MODS_PREPROCESS,
    RADIUS_CONF_TEMPLATE_SITES_DEFAULT,
]
RADIUS_CONF_TEMPLATES_PERMIT = [
    RADIUS_CONF_TEMPLATE_ROOT_AUTHORIZED_MACS,
    RADIUS_CONF_TEMPLATE_ROOT_USERS,
]

RADIUS_CONF_TEMPLATE_CERTS_CA_CNF = RADIUS_CONF_TEMPLATE_DIR + "/certs.ca.cnf.tpl"
RADIUS_CONF_TEMPLATE_CERTS_SERVER_CNF = RADIUS_CONF_TEMPLATE_DIR + "/certs.server.cnf.tpl"
RADIUS_CONF_TEMPLATE_CERTS_CLIENT_CNF = RADIUS_CONF_TEMPLATE_DIR + "/certs.client.cnf.tpl"

# ========================================================
# temporary freeradius configuration
# ========================================================
RADIUS_CONF_TEMPORARY_DIR = RADIUS_CONF_DIR + "/temporary"
RADIUS_CONF_TEMPORARY_ROOT_RADIUSD_CONF = RADIUS_CONF_TEMPORARY_DIR + "/root.radiusd.conf.tmp"
RADIUS_CONF_TEMPORARY_ROOT_CLIENTS_CONF = RADIUS_CONF_TEMPORARY_DIR + "/root.clients.conf.tmp"
RADIUS_CONF_TEMPORARY_ROOT_AUTHORIZED_MACS = RADIUS_CONF_TEMPORARY_DIR + "/root.authorized_macs.tmp"
RADIUS_CONF_TEMPORARY_ROOT_USERS = RADIUS_CONF_TEMPORARY_DIR + "/root.users.tmp"
RADIUS_CONF_TEMPORARY_MODS_EAP = RADIUS_CONF_TEMPORARY_DIR + "/mods.eap.tmp"
RADIUS_CONF_TEMPORARY_MODS_LOGINTIME = RADIUS_CONF_TEMPORARY_DIR + "/mods.logintime.tmp"
RADIUS_CONF_TEMPORARY_MODS_MSCHAP = RADIUS_CONF_TEMPORARY_DIR + "/mods.mschap.tmp"
RADIUS_CONF_TEMPORARY_MODS_PREPROCESS = RADIUS_CONF_TEMPORARY_DIR + "/mods.preprocess.tmp"
RADIUS_CONF_TEMPORARY_SITES_DEFAULT = RADIUS_CONF_TEMPORARY_DIR + "/sites.default.tmp"

RADIUS_CONF_TEMPORARYS = {
    RADIUS_CONF_TEMPLATE_ROOT_RADIUSD_CONF: RADIUS_CONF_TEMPORARY_ROOT_RADIUSD_CONF,
    RADIUS_CONF_TEMPLATE_ROOT_CLIENTS_CONF: RADIUS_CONF_TEMPORARY_ROOT_CLIENTS_CONF,
    RADIUS_CONF_TEMPLATE_ROOT_AUTHORIZED_MACS: RADIUS_CONF_TEMPORARY_ROOT_AUTHORIZED_MACS,
    RADIUS_CONF_TEMPLATE_ROOT_USERS: RADIUS_CONF_TEMPORARY_ROOT_USERS,
    RADIUS_CONF_TEMPLATE_MODS_EAP: RADIUS_CONF_TEMPORARY_MODS_EAP,
    RADIUS_CONF_TEMPLATE_MODS_LOGINTIME: RADIUS_CONF_TEMPORARY_MODS_LOGINTIME,
    RADIUS_CONF_TEMPLATE_MODS_MSCHAP: RADIUS_CONF_TEMPORARY_MODS_MSCHAP,
    RADIUS_CONF_TEMPLATE_MODS_PREPROCESS: RADIUS_CONF_TEMPORARY_MODS_PREPROCESS,
    RADIUS_CONF_TEMPLATE_SITES_DEFAULT: RADIUS_CONF_TEMPORARY_SITES_DEFAULT,
}

RADIUS_CONF_TEMPORARY_CERTS_CA_CNF = RADIUS_CONF_TEMPORARY_DIR + "/certs.ca.cnf.tmp"
RADIUS_CONF_TEMPORARY_CERTS_SERVER_CNF = RADIUS_CONF_TEMPORARY_DIR + "/certs.server.cnf.tmp"
RADIUS_CONF_TEMPORARY_CERTS_CLIENT_CNF = RADIUS_CONF_TEMPORARY_DIR + "/certs.client.cnf.tmp"

# ========================================================
# real freeradius configuration
# ========================================================
RADDB_DIR = "/mnt/data/etc/freeradius"
RADDB_CERTS_DIR = RADDB_DIR + "/certs"
RADDB_MODS_DIR = RADDB_DIR + "/modules"
RADDB_SITES_DIR = RADDB_DIR + "/sites-enabled"

RADDB_ROOT_RADIUSD_CONF = RADDB_DIR + "/radiusd.conf"
RADDB_ROOT_CLIENTS_CONF = RADDB_DIR + "/clients.conf"
RADDB_ROOT_AUTHORIZED_MACS = RADDB_DIR + "/authorized_macs"
RADDB_ROOT_USERS = RADDB_DIR + "/users"
RADDB_MODS_EAP = RADDB_DIR + "/eap.conf"
RADDB_MODS_LOGINTIME = RADDB_MODS_DIR + "/logintime"
RADDB_MODS_MSCHAP = RADDB_MODS_DIR + "/mschap"
RADDB_MODS_PREPROCESS = RADDB_MODS_DIR + "/preprocess"
RADDB_SITES_DEFAULT = RADDB_SITES_DIR + "/default"

RADDB_CONFS = {
    RADIUS_CONF_TEMPORARY_ROOT_RADIUSD_CONF: RADDB_ROOT_RADIUSD_CONF,
    RADIUS_CONF_TEMPORARY_ROOT_CLIENTS_CONF: RADDB_ROOT_CLIENTS_CONF,
    RADIUS_CONF_TEMPORARY_ROOT_AUTHORIZED_MACS: RADDB_ROOT_AUTHORIZED_MACS,
    RADIUS_CONF_TEMPORARY_ROOT_USERS: RADDB_ROOT_USERS,
    RADIUS_CONF_TEMPORARY_MODS_EAP: RADDB_MODS_EAP,
    RADIUS_CONF_TEMPORARY_MODS_LOGINTIME: RADDB_MODS_LOGINTIME,
    RADIUS_CONF_TEMPORARY_MODS_MSCHAP: RADDB_MODS_MSCHAP,
    RADIUS_CONF_TEMPORARY_MODS_PREPROCESS: RADDB_MODS_PREPROCESS,
    RADIUS_CONF_TEMPORARY_SITES_DEFAULT: RADDB_SITES_DEFAULT,
}

RADDB_CERTS_CA_CNF = RADDB_CERTS_DIR + "ca.cnf"
RADDB_CERTS_SERVER_CNF = RADDB_CERTS_DIR + "server.cnf"
RADDB_CERTS_CLIENT_CNF = RADDB_CERTS_DIR + "client.cnf"
RADDB_CERTS_CLIENTS_DIR = RADDB_CERTS_DIR + "/clients"


# ========================================================
# CertGenerator
# ========================================================
class CertGenerator(object):
    def __init__(self, conf_dict):
        self.cert_dir = RADDB_CERTS_DIR
        self.conf_dict = conf_dict
        conf_dict["config"]["certs"]["dir"] = RADDB_CERTS_DIR
        self.ca_passphrase = self.conf_dict["config"]["certs"]["ca"]["passphrase"]

    @property
    def path_ca_cnf(self):
        return os.path.join(self.cert_dir, "ca.cnf")

    @property
    def path_ca_key(self):
        return os.path.join(self.cert_dir, "ca.key")

    @property
    def path_ca_pem(self):
        return os.path.join(self.cert_dir, "ca.pem")

    @property
    def path_ca_der(self):
        return os.path.join(self.cert_dir, "ca.der")

    @property
    def path_crl_pem(self):
        return os.path.join(self.cert_dir, "crl.pem")

    @property
    def path_cacrl_pem(self):
        return os.path.join(self.cert_dir, "cacrl.pem")

    @property
    def path_server_cnf(self):
        return os.path.join(self.cert_dir, "server.cnf")

    @property
    def path_server_csr(self):
        return os.path.join(self.cert_dir, "server.csr")

    @property
    def path_server_key(self):
        return os.path.join(self.cert_dir, "server.key")

    @property
    def path_server_cnf(self):
        return os.path.join(self.cert_dir, "server.cnf")

    @property
    def path_server_crt(self):
        return os.path.join(self.cert_dir, "server.crt")

    @property
    def path_server_p12(self):
        return os.path.join(self.cert_dir, "server.p12")

    @property
    def path_server_pem(self):
        return os.path.join(self.cert_dir, "server.pem")

    @property
    def path_server_cnf(self):
        return os.path.join(self.cert_dir, "server.cnf")

    def client_file_name(self, user_name):
        return get_valid_filename(user_name)

    def path_client_csr(self, user_name):
        return os.path.join(RADDB_CERTS_CLIENTS_DIR, self.client_file_name(user_name) + ".csr")

    def path_client_key(self, user_name):
        return os.path.join(RADDB_CERTS_CLIENTS_DIR, self.client_file_name(user_name) + ".key")

    def path_client_cnf(self, user_name):
        return os.path.join(RADDB_CERTS_CLIENTS_DIR, self.client_file_name(user_name) + ".cnf")

    def path_client_crt(self, user_name):
        return os.path.join(RADDB_CERTS_CLIENTS_DIR, self.client_file_name(user_name) + ".crt")

    def path_client_p12(self, user_name):
        return os.path.join(RADDB_CERTS_CLIENTS_DIR, self.client_file_name(user_name) + ".p12")

    def path_client_pem(self, user_name):
        return os.path.join(RADDB_CERTS_CLIENTS_DIR, self.client_file_name(user_name) + ".pem")

    @property
    def path_xpextensions(self):
        return os.path.join(self.cert_dir, "xpextensions")

    @staticmethod
    def _exec(cmd, cwd='/mnt/data/etc/freeradius/certs', raise_exc=True):
        ret = ""
        proc = CmdExec(cmd.split(), stdin=NULL, stdout=PIPE, stderr=PIPE,
                       cwd=cwd,
                       raise_exc=False, close_fds=False).execute()
        output, err = proc.communicate()
        ret = {"cmd": cmd, "returncode": proc.returncode,  "stdout": output, "stderr": err}

        if proc.wait() != 0 and raise_exc:
            raise ExecError(ret)
        return ret

    def clear(self, initialize):
        """clear all certificate"""

        for target in (
            "*~",
            "*.csr",
            "*.crt",
            "*.p12",
            "*.der",
            "*.pem",
            "*.key",
            "clients/*.csr",
            "clients/*.crt",
            "clients/*.p12",
            "clients/*.der",
            "clients/*.pem",
            "clients/*.key",
            "index.txt*",
            "serial*",
            "*.[0-9]",
            "*.r0",
        ):
            _path1 = os.path.join(self.cert_dir, target)
            for _path2 in glob.glob(_path1):
                os.remove(_path2)
                # print(_path2)

        if initialize:
            self.initialize()

    def initialize(self):
        """generate index.txt and serial"""
        _path = os.path.join(self.cert_dir, "index.txt")
        with open(_path, "w") as fp:
            fp.write("")

        _path = os.path.join(self.cert_dir, "serial")
        with open(_path, "w") as fp:
            fp.write("00")

    def generate_ca_cert(self):
        # generate ca.cnf
        _dict = deepcopy(self.conf_dict)
        context = Context(_dict)
        with open(RADIUS_CONF_TEMPLATE_CERTS_CA_CNF) as fp:
            temp_src = fp.read()
        temp = Template(temp_src)
        dest = temp.render(context)
        with open(RADDB_CERTS_CA_CNF, "w") as fp:
            fp.write(dest)

        # generate CA cert
        try:
            cmd = "{0} req -new -x509 -keyout {1} -out {2} -config {3} -passin pass:{4} -passout pass:{5}".format(
                OPENSSL_CMD,
                self.path_ca_key, self.path_ca_pem, self.path_ca_cnf,
                self.ca_passphrase, self.ca_passphrase,
            )
            self._exec(cmd)
            # convert DER
            cmd = "{0} x509 -inform PEM -outform DER -in {1} -out {2}".format(
                OPENSSL_CMD,
                self.path_ca_pem, self.path_ca_der,
            )
            self._exec(cmd)
        except Exception:
            for f in (self.path_ca_key, self.path_ca_pem):
                if os.path.exists(f):
                    os.remove(f)
            raise

    def extend_ca_cert(self, days):
        cmd = "{0} req -x509 -days {1} -key {2} -out {3} -passin pass:{4} -config {5}".format(
            OPENSSL_CMD,
            days, self.path_ca_key, self.path_ca_pem, self.ca_passphrase, self.path_ca_cnf
        )
        self._exec(cmd)

    def get_expiration(self, pem):
        cmd = "{0} x509 -in {1} -noout -enddate".format(OPENSSL_CMD, pem)
        result = self._exec(cmd, raise_exc=False)
        if result["returncode"] != 0:
            return None
        else:
            txt = result["stdout"]
            dt = parser.parse(txt[9:-1], )
            dt = dt.astimezone(tz.tzlocal())
            # return dt.strftime("%Y/%m/%d %H:%M:%S %Z")
            return dt

    @property
    def ca_expiration(self):
        return self.get_expiration(self.path_ca_pem)

    def generate_ca_crl(self):
        try:
            cmd = "{0} ca -gencrl -cert {1} -keyfile {2} -config {3} -passin pass:{4} -out {5}".format(
                OPENSSL_CMD,
                self.path_ca_pem,
                self.path_ca_key,
                self.path_ca_cnf,
                self.ca_passphrase,
                self.path_crl_pem,
            )
            self._exec(cmd)

            # rehash
            cmd = "{0} {1}".format(
                C_REHASH_CMD,
                self.cert_dir
            )
            self._exec(cmd)
            
        except Exception:
            for f in (self.path_crl_pem, ):
                if os.path.exists(f):
                    os.remove(f)
            raise

    def extend_ca_crl(self, days):
        cmd = "{0} ca -gencrl -crldays{1} -config {2} -out {3} -passin pass:{4}".format(
            OPENSSL_CMD,
            days, self.path_ca_cnf, self.path_crl_pem, self.ca_passphrase
        )
        self._exec(cmd)

        # rehash
        cmd = "{0} {1}".format(
            C_REHASH_CMD,
            self.cert_dir
        )
        self._exec(cmd)
        
    @property
    def crl_expiration(self):
        return self.get_expiration(self.path_crl_pem)

    def generate_server_cert(self, days=None):
        # generate server.cnf
        _dict = deepcopy(self.conf_dict)
        context = Context(_dict)
        with open(RADIUS_CONF_TEMPLATE_CERTS_SERVER_CNF) as fp:
            temp_src = fp.read()
        temp = Template(temp_src)
        dest = temp.render(context)
        with open(RADDB_CERTS_SERVER_CNF, "w") as fp:
            fp.write(dest)

        str_days = ""
        if days:
            str_days = "-days {0}".format(days)

        try:
            # generate CSR
            cmd = "{0} req -new -keyout {1} -out {2} -config {3} -passin pass:{4}  -passout pass:{5}".format(
                OPENSSL_CMD,
                self.path_server_key, self.path_server_csr, self.path_server_cnf,
                self.ca_passphrase, self.ca_passphrase,
            )
            self._exec(cmd)
            # sign by CA
            tmp = "{0} ca -batch {1} -keyfile {2} -key {3} -cert {4} -in {5} -out {6} " \
                  "-extensions xpserver_ext -extfile {7} -config {8}"
            cmd = tmp.format(
                OPENSSL_CMD,
                str_days,
                self.path_ca_key, self.ca_passphrase, self.path_ca_pem,
                self.path_server_csr, self.path_server_crt, self.path_xpextensions, self.path_server_cnf,
            )
            self._exec(cmd)
            # convert PKCS#12
            cmd = "{0} pkcs12 -export -in {1} -inkey {2} -out {3} -passin pass:{4}  -passout pass:{5}".format(
                OPENSSL_CMD,
                self.path_server_crt, self.path_server_key, self.path_server_p12,
                self.ca_passphrase, self.ca_passphrase,
            )
            self._exec(cmd)
            # convert PEM
            cmd = "{0} pkcs12 -in {1} -out {2} -passin pass:{3}  -passout pass:{4}".format(
                OPENSSL_CMD,
                self.path_server_p12, self.path_server_pem,
                self.ca_passphrase, self.ca_passphrase
            )
            self._exec(cmd)

            # rehash
            cmd = "{0} {1}".format(
                C_REHASH_CMD,
                self.cert_dir
            )
            self._exec(cmd)

        except Exception:
            for f in (self.path_server_key, self.path_server_csr, self.path_server_crt,
                      self.path_server_pem, self.path_server_p12):
                if os.path.exists(f):
                    os.remove(f)
            raise

    def revoke_server_cert(self):
        try:
            # revoke cert & output crl
            cmd = "{0} ca -revoke {1} -cert {2} -keyfile {3} -passin pass:{4} -config {5} -out {6}".format(
                OPENSSL_CMD,
                self.path_server_pem, self.path_ca_pem, self.path_ca_key, self.ca_passphrase,
                self.path_ca_cnf, self.path_crl_pem
            )
            self._exec(cmd)

            # rehash
            cmd = "{0} {1}".format(
                C_REHASH_CMD,
                self.cert_dir
            )
            self._exec(cmd)
            
        except Exception:
            raise

    def extend_server_cert(self, days):
        # revoke
        self.revoke_server_cert()

        # generate
        self.generate_server_cert(days)

    def verify_server_cert(self):
        cmd = "{0} verify -CAfile {1} {2}".format(
            OPENSSL_CMD,
            self.path_ca_pem, self.path_server_pem
        )
        results = self._exec(cmd, raise_exc=False)
        return results["returncode"] == 0

    @property
    def server_expiration(self):
        return self.get_expiration(self.path_server_pem)

    def generate_client_cert(self, user_name, user_pass):
        cnf = self.path_client_cnf(user_name)
        key = self.path_client_key(user_name)
        csr = self.path_client_csr(user_name)
        crt = self.path_client_crt(user_name)
        p12 = self.path_client_p12(user_name)
        pem = self.path_client_pem(user_name)
        if not os.path.exists(RADDB_CERTS_CLIENTS_DIR):
            os.mkdir(RADDB_CERTS_CLIENTS_DIR)

        # generate client.cnf
        _dict = deepcopy(self.conf_dict)
        _dict["config"]["certs"]["client"] = {
            "commonName": user_name,
            "emailAddress": "dummy",
            "passphrase": user_pass,
        }
        context = Context(_dict)
        with open(RADIUS_CONF_TEMPLATE_CERTS_CLIENT_CNF) as fp:
            temp_src = fp.read()
        temp = Template(temp_src)
        dest = temp.render(context)
        with open(cnf, "w") as fp:
            fp.write(dest)

        try:
            user_pass2 = ""
            if user_pass != "":
                user_pass2 = " -passin pass:{0} -passout pass:{1}".format(user_pass, user_pass)
                user_pass3 = " -passin pass:{0} -passout pass:{1}".format(user_pass, user_pass)
                user_pass4 = " -passin pass:{0} -passout pass:{1}".format(user_pass, user_pass)
            else:
                user_pass2 = " -passin pass:{0} -passout pass:{1}".format("dummy", "dummy")
                user_pass3 = " -passin pass:{0} -passout pass:{1}".format("dummy", "")
                user_pass4 = " -passin pass:{0} -nodes".format("")
                
            # generate CSR
            cmd = "{0} req -new -keyout {1} -out {2} -config {3} {4}".format(
                OPENSSL_CMD,
                key, csr, cnf, user_pass2,
            )
            self._exec(cmd)
            
            # sign by CA
            tmp = "{0} ca -batch -keyfile {1} -cert {2} -key {3} -in {4} -out {5} "\
                  "-extensions xpclient_ext -extfile {6} -config {7}"
            cmd = tmp.format(
                OPENSSL_CMD,
                self.path_ca_key, self.path_ca_pem, self.ca_passphrase,
                csr, crt, self.path_xpextensions, cnf,
            )
            self._exec(cmd)
            
            # convert PKCS#12
            cmd = "{0} pkcs12 -export -in {1} -inkey {2} -out {3} {4}".format(
                OPENSSL_CMD,
                crt, key, p12, user_pass3, 
            )
            self._exec(cmd)
            # convert PEM
            cmd = "{0} pkcs12 -in {1} -out {2} {3}".format(
                OPENSSL_CMD,
                p12, pem, user_pass4,
            )
            self._exec(cmd)

            # rehash
            cmd = "{0} {1}".format(
                C_REHASH_CMD,
                self.cert_dir
            )
            self._exec(cmd)

        except Exception:
            #for f in (cnf, key, csr, crt, p12, pem):
            #    if os.path.exists(f):
            #        os.remove(f)
            raise

    def revoke_client_cert(self, user_name):
        cnf = self.path_client_cnf(user_name)
        key = self.path_client_key(user_name)
        csr = self.path_client_csr(user_name)
        crt = self.path_client_crt(user_name)
        p12 = self.path_client_p12(user_name)
        pem = self.path_client_pem(user_name)
        try:
            # revoke cert
            cmd = "{0} ca -revoke {1} -cert {2} -keyfile {3} -passin pass:{4} -config {5}".format(
                OPENSSL_CMD,
                pem, self.path_ca_pem, self.path_ca_key, self.ca_passphrase,
                self.path_ca_cnf
            )
            self._exec(cmd)
            
            #  output crl
            cmd = "{0} ca -gencrl -passin pass:{1} -config {2} -out {3}".format(
                OPENSSL_CMD,
                self.ca_passphrase,
                self.path_ca_cnf, self.path_crl_pem
            )
            self._exec(cmd)

            # rehash
            cmd = "{0} {1}".format(
                C_REHASH_CMD,
                self.cert_dir
            )
            self._exec(cmd)
            
            # remove cert
            for f in (cnf, key, csr, crt, p12, pem):
                if os.path.exists(f):
                    os.remove(f)
        except Exception:
            raise

    def extend_client_cert(self, user_name, days):
        # revoke
        self.revoke_client_cert(user_name)

        # generate
        self.generate_client_cert(user_name, days)

    def verify_client_cert(self, user_name):
        file_name = get_valid_filename(user_name)
        pem = os.path.join(RADDB_CERTS_CLIENTS_DIR, file_name + '.pem')
        cmd = "{0} verify -CAfile {1} {2}".format(
            OPENSSL_CMD,
            self.path_ca_pem, pem
        )
        self._exec(cmd, raise_exc=False)

    def client_expiration(self, user_name):
        file_name = get_valid_filename(user_name)
        pem = os.path.join(RADDB_CERTS_CLIENTS_DIR, file_name + '.pem')
        return self.get_expiration(pem)

    def get_client_list(self):
        ret = []
        index_path = os.path.join(self.cert_dir, "index.txt")
        with open(index_path) as fp:
            tsv = csv.reader(fp, delimiter='\t')
            pattern = re.compile(r'CN=(.+)$')
            for row in tsv:
                if row[0] == "R":
                    # revoke
                    continue

                tmp = row[5]
                for col in tmp.split("/"):
                    m = re.search(pattern, col)
                    if m:
                        name = m.group(1)
                        if name != SERVER_COMMON_NAME:
                            ret.append(m.group(1))
                        break
        return ret


# ========================================================
# PalRadius
# ========================================================
class PalRadius(palcmd.PalCmd):
    """Pal Radius command implementation class."""

    # Default Settings
    default_settings = {
        "config": {
            "default_eap_type": "reject",
            "method": {
                "macaddr": True,
                "peap": True,
                "eaptls": True,
            },
            "radiusd": {
                "max_request_time": 30,
                "cleanup_delay": 5,
                "security": {
                    "status_server": "yes",
                    "max_attributes": 200,
                    "reject_delay": 1
                },
                "max_requests": 16384,
                "thread_pool": {
                    "start_servers": 5,
                    "max_servers": 32,
                    "max_requests_per_server": 0,
                    "max_queue_size": 65536,
                    "min_spare_servers": 3,
                    "max_spare_servers": 10
                }
            },
            "mods": {
                "preprocess": {
                    "ascend_channels_per_line": 23,
                    "with_ascend_hack": "no",
                    "with_specialix_jetstream_hack": "no",
                    "with_ntdomain_hack": "no",
                    "with_cisco_vsa_hack": "no"
                },
                "eap": {
                    "mschapv2": {
                        "send_error": "no",
                        "identity": "\"FreeRADIUS\""
                    },
                    "timer_expire": 60,
                    "cisco_accounting_username_bug": "no",
                    "peap": {
                        "require_client_cert": "no",
                        "copy_request_to_tunnel": "no",
                        "default_eap_type": "mschapv2"
                    },
                    "tls_common": {
                        "tls_max_version": "\"1.2\"",
                        "cipher_list": "\"DEFAULT\"",
                        "check_crl": "yes",
                        "auto_chain": "yes",
                        "disable_tlsv1_2": "no",
                        "disable_tlsv1_1": "no",
                        "cipher_server_preference": "no",
                        "disable_tlsv1": "no",
                        "check_all_crl": "yes",
                        "fragment_size": 1024,
                        "cache": {
                            "lifetime": 24,
                            "enable": "no",
                            "name": "\"EAP module\""
                        },
                        "ecdh_curve": "\"prime256v1\"",
                        "tls_min_version": "\"1.0\"",
                        "allow_expired_crl": "no",
                        "include_length": "yes"
                    }
                },
                "mschap": {
                    "require_strong": "yes",
                    "use_open_directory": "yes",
                    "retry_msg": "\"Re-enter (or reset) the password\"",
                    "use_mppe": "yes",
                    "require_encryption": "yes",
                    "allow_retry": "yes",
                    "ntlm_auth_timeout": 10,
                },
                "logintime": {
                    "minimum_timeout": 60
                }
            },
            "sites": {
                "default": {
                    "listen": {
                        "auth": {
                            "v4": {
                                "port": 1812,
                                "limit": {
                                    "lifetime": 0,
                                    "idle_timeout": 30,
                                    "max_connections": 16
                                },
                            },
                            "v6": {
                                "port": 1812,
                                "limit": {
                                    "lifetime": 0,
                                    "idle_timeout": 30,
                                    "max_connections": 16
                                },
                            }
                        }
                    }
                }
            },
            "certs": {
                "dir": RADDB_CERTS_DIR,
                "ca": {
                    "countryName": "JP",
                    "stateOrProvinceName": None,
                    "localityName": None,
                    "organizationName": None,
                    "commonName": CA_COMMON_NAME,
                    "emailAddress": None,
                    "passphrase": "NL65cmWkjPT",
                },
                "server": {
                    "commonName": SERVER_COMMON_NAME,
                    "emailAddress": None,
                    "passphrase": "kZwMmLft1r7",
                }
            },
        },
        "clients": [],
        "users": [],
        "macs": [],
    }

    #
    # Internal methods
    #

    def _check_body_node(self, node, type=None):
        """Check if entry's body node is valid."""

        if node.get_node_name() != RADIUS_NAME:
            raise InvalidArgError(node.get_node_name())

    def _nt_hash(self, clear_text):
        hash = hashlib.new('md4', clear_text.encode('utf-16le')).digest()
        return binascii.hexlify(hash)

    def _get_client_cert_list(self):
        cg = CertGenerator(self.default_settings)
        return cg.get_client_list()
    
    def _add_client_cert_list_to_entry(self, entry, cert_list):
        """ add certificatelist to entry """
        radius_node = entry.get()
        radius_node.append_child("certificatelist")
        list_node = radius_node.get_node("certificatelist")

        for client in self._get_client_cert_list():
            item_node = XmlNode("certificate")
            item_node.append_child("name", client)
            list_node.append_child(item_node)

    def _delete_client_cert(self, entry):
        """ revoke client certificate """
        user_name = entry.get().get_node("certificate/name").get_text()
        cg = CertGenerator(self.default_settings)
        cg.revoke_client_cert(user_name)

    def _deploy_client_cert(self, entry):
        dest = entry.get_file()
        user_name = entry.get().get_node("certificate/name").get_text()
        
        if user_name in self._get_client_cert_list():
            cg = CertGenerator(self.default_settings)
            src = cg.path_client_p12(user_name)
            shutil.copy(src, dest)
        else:
            raise LookupError("not found(%s)" % user_name)

    def _create_client_cert(self, entry):
        file_path = entry.get_file()
        user_name = entry.get().get_node("certificate/name").get_text()
        user_pass = entry.get().get_node("certificate/passphrase").get_text()
        
        if user_name in self._get_client_cert_list():
            raise LookupError("duplicate(%s)" % user_name)
        else:
            cg = CertGenerator(self.default_settings)
            cg.generate_client_cert(user_name, user_pass)

    def _build_temporary_authenticate(self, entry):
        """build config files to temporary dir"""

        # Update Settings
        entry_root = entry.get()

        try:
            authenticate_root = entry_root.get_node("./authenticate")
        except NotFoundError:
            return

        # todo
        context_dict = deepcopy(self.default_settings)
        context_dict["config"]["sites"]["default"]["listen"]["auth"]["v4"]["port"] = entry_root.get_node("./authenticate/port").get_text()
        context_dict["config"]["sites"]["default"]["listen"]["auth"]["v6"]["port"] = entry_root.get_node("./authenticate/port").get_text()
        
        for method in ["macaddr", "peap", "eaptls", "adauth"]:
            context_dict["config"]["method"][method] = entry_root.get_node("./authenticate/method").is_set_to_true(method)

        if authenticate_root.is_enabled():
            if context_dict["config"]["method"]["peap"]:
                context_dict["config"]["default_eap_type"] = "peap"
            else:
                if context_dict["config"]["method"]["eaptls"]:
                    context_dict["config"]["default_eap_type"] = "tls"
                else:
                    if context_dict["config"]["method"]["adauth"]:
                        context_dict["config"]["default_eap_type"] = "mschapv2"
                    else:
                        context_dict["config"]["default_eap_type"] = "reject"

        context_dict["clients"] = []
        for authenticator in authenticate_root.get_node_list("./authenticatorlist/authenticator"):
            name = authenticator.get_node("./name").get_text()
            ipaddr = authenticator.get_node("./ipaddr").get_text()
            secret = authenticator.get_node("./secretkey").get_text()
            context_dict["clients"].append({
                "name": name,
                "ipv4addr": ipaddr,
                "ipv6addr": "",
                "secret": secret,
                "nas_type": "other",
            })

        self._build_temporary_files(context_dict, RADIUS_CONF_TEMPLATES_AUTHENTICATE)

    def _build_temporary_permit(self, entry):
        """build config files to temporary dir"""

        # Update Settings
        entry_root = entry.get()
        try:
            permit_root = entry_root.get_node("./permit")
        except NotFoundError:
            return

        context_dict = deepcopy(self.default_settings)

        context_dict["users"] = []
        for user in permit_root.get_node_list("./userlist/user"):
            username = user.get_node("./name").get_text()

            password_node = user.get_node("./password")
            userpass = password_node.get_text()
            if not password_node.is_set_to_true("hashed"):
                userpass = self._nt_hash(userpass)
                password_node.set_text(userpass)
                password_node.set_bool_attr("hashed", True)

            context_dict["users"].append({
                "name": username,
                "secret": userpass,
            })

        context_dict["macs"] = []
        for macaddr in permit_root.get_node_list("./macaddrlist/macaddr"):
            addr = macaddr.get_text()
            addr = addr.replace(":", "-").upper()
            context_dict["macs"].append({
                "addr": addr,
            })

        self._build_temporary_files(context_dict, RADIUS_CONF_TEMPLATES_PERMIT)

    def _build_temporary_files(self, context_dict, templates):
        context = Context(context_dict)
        for temp_path in templates:
            with open(temp_path) as fp:
                temp_src = fp.read()
            temp = Template(temp_src)
            dest = temp.render(context)

            output_path = RADIUS_CONF_TEMPORARYS[temp_path]
            with open(output_path, "w") as fp:
                fp.write(dest)

    def _deploy_conf(self):
        """move config files from temporary dir to raddb dir"""
        for template in RADIUS_CONF_TEMPORARYS:
            temporary = RADIUS_CONF_TEMPORARYS[template]
            conf_path = RADDB_CONFS[temporary]

            if not os.path.exists(temporary):
                continue

            shutil.copy(temporary, conf_path)

    @staticmethod
    def _exec(cmd, raise_exc=True):
        ret = ""
        proc = CmdExec(cmd.split(), stdin=NULL, stdout=PIPE, stderr=PIPE,
                       raise_exc=False, close_fds=False).execute()
        output, err = proc.communicate()
        ret = {"cmd": cmd, "returncode": proc.returncode,  "stdout": output, "stderr": err}

        if proc.wait() != 0 and raise_exc:
            raise ExecError(ret)
        return ret

    #
    # Request handlers
    #
    def _list_handler(self, entry):
        """Handle list request."""

        self._check_body_node(entry.get())

        try:
        	cert_list = self._get_client_cert_list()
            self._add_client_cert_list_to_entry(entry, cert_list)
        except:
            self.exception("list fail.")
            entry.set_result(Results.LIST_ERROR)
            return entry

        entry.set_result(Results.OK)
        return entry

    def _delete_handler(self, entry):
        """Handle delete request."""

        self._check_body_node(entry.get())

        try:
            self._delete_client_cert(entry)
        except:
            self.exception("delete fail.")
            entry.set_result(Results.DELETE_ERROR)
            return entry

        entry.set_result(Results.OK)
        return entry

    def _get_handler(self, entry):
        """Handle get request."""

        self._check_body_node(entry.get())

        try:
            self._deploy_client_cert(entry)
        except:
            self.exception("get fail.")
            entry.set_result(Results.GET_ERROR)
            return entry

        entry.set_result(Results.OK)
        return entry

    def _create_handler(self, entry):
        """Handle create request."""

        self._check_body_node(entry.get())

        try:
            self._create_client_cert(entry)
            self._deploy_client_cert(entry)
        except:
            self.exception("create fail.")
            entry.set_result(Results.CREATE_ERROR)
            return entry

        entry.set_result(Results.OK)
        return entry

    def _apply_handler(self, entry):
        """Handle apply request."""

        self._check_body_node(entry.get())

        try:
            self._build_temporary_authenticate(entry)
            self._build_temporary_permit(entry)
            self._deploy_conf()
        except:
            self.exception("apply fail.")
            entry.set_result(Results.APPLY_ERROR)
            return entry

        entry.set_result(Results.OK)
        return entry

    #
    # Constructor
    #

    def __init__(self, option=None):
        """Constructor."""

        palcmd.PalCmd.__init__(self, option=option)

        self.set_handler(ReqTypes.APPLY, self._apply_handler)
        self.set_handler(ReqTypes.CREATE, self._create_handler)
        self.set_handler(ReqTypes.GET, self._get_handler)
        self.set_handler(ReqTypes.DELETE, self._delete_handler)
        self.set_handler(ReqTypes.LIST, self._list_handler)


#
# Main
#

if __name__ == "__main__":
    PalRadius(option=palcmd.PalCmdOption()).handle_request()
