#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""Do something for installing the hradius package."""

import os
import sys

from iobase import xmldata
from iobase import cmdexec
from ionas import request
from product_conf import (BIN_DIR, PACKAGES_DIR, XML_CONF_DIR,)
from fal_info import (FAL_SERVICECTRL_PATH, Results, ReqTypes)

import fal_msginfo

from pkgaddon.hradius_config import (RADIUS_SERVICE,)

import manualctrl


ENCODING = "UTF-8"
CATEGORYLIST_PATH = "categorylist"
CATEGORY_NAME = "category"
CATEGORY_KEY_PATH = "code"
CATEGORY_FIND_XPATH_FMT = """category[code="%s"]"""

MESSAGEINFOLIST_PATH = "messageinfolist"
MESSAGEINFO_NAME = "messageinfo"
MESSAGEINFO_KEY_PATH = "code"
MESSAGEINFO_FIND_XPATH_FMT = """messageinfo[code="%s"]"""

MESSAGES_FILE = os.path.join(XML_CONF_DIR, "messages.xml")
PACKAGE_MESSAGES_FILE = os.path.join(PACKAGES_DIR, "hradius", "messages.xml")
BUILD_MSGINFO_CMD = os.path.join(BIN_DIR, "build_msginfo")

PYC_SUFFIX = ".pyc"
PY_SUFFIX = ".py"

SERVICELIST_NAME = "servicelist"
RADIUS_NAME_XPATH = SERVICELIST_NAME + \
                    '/service/name[text()="%s"]' % (RADIUS_SERVICE,)
RADIUS_SERVICE_XML = \
        '<service enable="true"><name>%s</name></service>' % (RADIUS_SERVICE,)
SERVICECTRL_REQ_XML = """<servicectrl>
    <servicelist>
        <service><name>%s</name></service>
    </servicelist>
</servicectrl>
""" % (RADIUS_SERVICE,)


def __send_request(target, type, body=""):
    """Send request entry to "target"."""

    entry = request.Entry(type, body)
    return request.Request(target).request(entry).get_entry()


def __merge_list(new_messages_xml, messages_xml, list_xpath,
        item_name, key_in_item_xpath, find_xpath_fmt):
    """Merge a list in "new_mssages_xml" into "messages_xml"."""

    new_list_node = new_messages_xml.get_node(list_xpath)
    list_node = messages_xml.get_node(list_xpath)

    for item_node in new_list_node.get_node_list(item_name):
        key = item_node.get_value(key_in_item_xpath)
        try:
            cur_node = list_node.get_node(find_xpath_fmt % (key,))
            if (cur_node.compare(item_node, ignore_order=True,
                                    ignore_indent=True)):
                cur_node.update(item_node, override=True)
        except:
            list_node.append_child(item_node.get_clone())


def __merge_categorylist(new_messages_xml, messages_xml):
    """Merge "categorylist" tree in "new_mssages_xml" into "messages_xml"."""

    __merge_list(new_messages_xml, messages_xml, CATEGORYLIST_PATH,
        CATEGORY_NAME, CATEGORY_KEY_PATH, CATEGORY_FIND_XPATH_FMT)


def __merge_messageinfolist(new_messages_xml, messages_xml):
    """Merge "messageinfolist" tree in "new_mssages_xml"
       into "messages_xml".
    """

    __merge_list(new_messages_xml, messages_xml, MESSAGEINFOLIST_PATH,
        MESSAGEINFO_NAME, MESSAGEINFO_KEY_PATH, MESSAGEINFO_FIND_XPATH_FMT)


def __update_msginfo(messages_file):
    """Update a "fal_msginfo" for "message_file"."""

    fal_msginfo_path = fal_msginfo.__file__

    (base, suffix) = os.path.splitext(fal_msginfo_path)
    if suffix == PYC_SUFFIX:
        fal_msginfo_path = base + PY_SUFFIX

    cmd_args = [BUILD_MSGINFO_CMD, messages_file, fal_msginfo_path]
    cmdexec.CmdExec(cmd_args).execute()


def __is_radius_service_registered(servicectrl_config):
    """Check if service for the radius package is already registered."""

    try:
        servicectrl_config.get_node(RADIUS_NAME_XPATH)
        return True
    except:
        return False


def update_messages(messages_file, package_messages_file):
    """Update "messages_file" by "package_messages_file"."""

    messages_xml = xmldata.XmlConf(messages_file, encoding=ENCODING)
    package_messages_xml = xmldata.XmlConf(package_messages_file,
                                encoding=ENCODING)

    __merge_categorylist(package_messages_xml, messages_xml)
    __merge_messageinfolist(package_messages_xml, messages_xml)
    tmp_path = messages_file + "." + str(os.getpid())
    messages_xml.save(tmp_path)
    os.rename(tmp_path, messages_file)

    __update_msginfo(messages_file)


def add_radius_service():
    """Add service for the radius package to a servicectrl's configuration
       file.
    """

    servicectrl_config = \
                    __send_request(FAL_SERVICECTRL_PATH, ReqTypes.LOAD).get()

    if __is_radius_service_registered(servicectrl_config):
        return

    service = xmldata.XmlConfNode(RADIUS_SERVICE_XML)
    servicectrl_config.get_node(SERVICELIST_NAME).append_child(service)
    if (__send_request(FAL_SERVICECTRL_PATH, ReqTypes.SAVE,
                servicectrl_config).get_result() != Results.OK):

        sys.exit(1)


def start_radius_service():
    """Start radius service."""

    __send_request(FAL_SERVICECTRL_PATH, ReqTypes.START, SERVICECTRL_REQ_XML)


#
# Main
#

update_messages(MESSAGES_FILE, PACKAGE_MESSAGES_FILE)

manualctrl.install()
add_radius_service()
start_radius_service()
