# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta


import os
import re
import settings

from django.utils.translation import ugettext as _
from django.core.exceptions import ValidationError
from django.core.validators import MaxValueValidator, MinValueValidator
from django.http import HttpResponseRedirect, HttpResponse
from django import forms
from django.forms import IntegerField
from django.forms import Field
from project.view import FuncView
from project.forms.fields.extras.enable import EnableField
from project.forms.fields.checkbox import CheckboxField
from project.forms.fields.extras.file import FileField
from ..radius import Radius

MAX_FILE_SIZE = 1000000

class ConfForm(forms.Form):

    def __init__(self, *args, **kwargs):
        super(ConfForm, self).__init__(*args, **kwargs)

        self.fields["radius_enable"] = EnableField(
                label=_(u'radius server enable'),
                initial=False)
        self.fields["radius_port"] = IntegerField(
                label=_(u"radius port number"),
                validators=[MinValueValidator(1), MaxValueValidator(65535)])
        self.fields["auth_mac"] = EnableField(
                label=_(u'mac authentication'),
                initial=False)
        self.fields["auth_peap"] = EnableField(
                label=_(u'peap authentication'),
                initial=False)
        self.fields["auth_adauth"] = CheckboxField(
                label=_(u'adauth authentication'))
        self.fields["auth_cert"] = EnableField(
                label=_(u'cert authentication'),
                initial=False)
        #self.fields["import_file"] = FileField(
        #        label=_(u'Import config File'),
        #        required=False)


class ConfView(FuncView):

    views = ['apply', 'export', 'import']
    radius = Radius()

    def __initial(self):
        try:
            self.radius.load()
        except:
            pass

        conf = self.radius.get_config()
        return {
            "radius_enable": conf["radius_enable"],
            "radius_port": conf["radius_port"],
            "auth_mac": conf["auth_mac"],
            "auth_peap": conf["auth_peap"],
            "auth_adauth": conf["auth_adauth"],
            "auth_cert": conf["auth_cert"],
        }


    def view(self, request):
        param = self.__initial()
        form = ConfForm(initial=param)

        return {
            self.KEY_TEMPLATE: os.path.join(self.TEMPLATES, "template.html"),
            self.KEY_TEMPLATE_DIC: {'form': form}
        }


    def view_apply(self, request):
        form = ConfForm(data=request.POST)

        with self.processing() as p:
            p.break_if_state_change()
            result = False
            if form.is_valid():
                conf = self.radius.get_config()
                conf['radius_enable'] = form.cleaned_data['radius_enable']
                conf['radius_port'] = form.cleaned_data['radius_port']
                conf['auth_mac'] = form.cleaned_data["auth_mac"]
                conf['auth_peap'] = form.cleaned_data["auth_peap"]
                conf['auth_adauth'] = form.cleaned_data["auth_adauth"]
                conf['auth_cert'] = form.cleaned_data["auth_cert"]
                try:
                    self.radius.save()
                    self.set_message(_(u'update configuration successful'))
                    result = True
                except Exception as e:
                    debug = e.args
                    self.set_message(_(u'update configuration failed'))
            p.result(result)

        return {
            self.KEY_TEMPLATE: os.path.join(self.TEMPLATES, "template.html"),
            self.KEY_TEMPLATE_DIC: {'form': form}
        }


    def view_export(self, request):
        try:
            filepath = self.radius.get_path_export_csv()
            filename = os.path.basename(filepath)
            if self.radius.export_csv_file(filepath) != Radius.Ok:
                raise Exception
            response = HttpResponse(
                open(filepath, 'rb').read(),
                mimetype="application/octet-stream")
            response['Content-Disposition'] = 'attachment; filename=%s' % filename
            return {self.KEY_RESPONSE: response}
        except:
            pass

        with self.processing() as p:
            p.break_if_state_change()
            self.set_message(_(u'export configuration failed'))
            p.result(False)
        return self.view(request)


    def view_import(self, request):
        return {self.KEY_REDIRECT: "/".join([settings.MENU_URL, "import"])}


class ImportForm(forms.Form):
    import_file = FileField(
        label=_(u'Import config File'),
        required=False)

    def __init__(self, *args, **kwargs):
        super(ImportForm, self).__init__(*args, **kwargs)

    def clean(self):
        file = self.cleaned_data.get("import_file")
        if not file:
            raise ValidationError(_(u'"Import config file" is required.'))
        return self.cleaned_data


class ImportView(FuncView):

    views = ['execute', 'cancel']

    def view(self, request):
        form = ImportForm()
        return {
            self.KEY_TEMPLATE: os.path.join(self.TEMPLATES, "import.html"),
            self.KEY_TEMPLATE_DIC: {"form": form}
        }

    def view_execute(self, request):
        form = ImportForm(data=request.POST, files=request.FILES)
        radius = Radius()
        filepath = radius.get_path_import_csv()

        try:
            if form.is_valid():
                file = form.cleaned_data["import_file"]
                if not file:
                    self.set_message(_(u'invalid import config'))
                elif file.size > MAX_FILE_SIZE:
                    self.set_message(_(u'Import config file too large'))
                else:
                    dest = open(filepath, "wb")
                    for chunk in file.chunks():
                        dest.write(chunk)
                    dest.close()
                    if radius.import_csv_file(filepath) == Radius.Ok:
                        self.set_message(_(u'Import config file is successful'))
                    else:
                        self.set_message(_(u'Import config file is failed'))
        except:
            self.set_message(_(u'Import config file is failed'))

        return {
            self.KEY_TEMPLATE: os.path.join(self.TEMPLATES, "import.html"),
            self.KEY_TEMPLATE_DIC: {'form': form}
        }


    def view_cancel(self, request):
        return {self.KEY_REDIRECT: "/".join([settings.MENU_URL, "config"])}

