# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import absolute_import
from __future__ import division

import os
import re
import sys
import json
import socket
import binascii

from django.utils.translation import ugettext as _
from iobase.exception import (NotFoundError, InvalidArgError,)
from iobase.xmldata import (XmlNode, XmlConfNode, escape)
from iobase.cmdexec import (CmdExec, NULL, PIPE, NotFoundError, ExecError, )
from ionas.request import (Request, Entry)
from product_conf import (SUDO_CMD, MOUNT_CMD, XML_CONF_DIR,)
from fal_info import (FAL_NOTICE_PATH, FAL_USERS_PATH, ReqTypes, Results)
from pkgaddon.fal_hradius_config import (CONFIG_FILE,)
from pkgaddon.pal_hradius_config import (RADIUS_CONF_DIR,)
from pkgaddon.fal_hradius_info import (FAL_RADIUS_PATH,)
from pkgaddon.pal_hradius_info import (PAL_RADIUS_PATH,)

sys.path.append(os.path.dirname(PAL_RADIUS_PATH))
from pal_radius import (RADIUS_CONF_TEMPORARY_DIR, RADDB_CERTS_DIR, SERVER_COMMON_NAME,)
#RADIUS_CONF_TEMPORARY_DIR = RADIUS_CONF_DIR + "/templates"
#RADDB_CERTS_DIR = "/mnt/data/etc/freeradius/certs"
#RADIUS_NAME = "landisk"

RADIUS_CERT_SERVER = RADDB_CERTS_DIR + "/server.crt"
RADIUS_CERT_CLIENT = RADIUS_CONF_TEMPORARY_DIR + "/%(name)s.p12"
RADIUS_CONF_FILE = XML_CONF_DIR + "/" + CONFIG_FILE
RADIUS_CSV_IMPORT = "/tmp/radius_import.csv"
RADIUS_CSV_EXPORT = "/tmp/radius_export.csv"

RADIUS_CLIENT_MAX = 100
RADIUS_MAC_MAX = 1000
RADIUS_USER_MAX = 1000

class Radius:
    Ok = 0
    Failed = -1
    ErrExistAlready = -2
    ErrNotFound = -3
    ErrBadFormat = -4
    ErrExceededUpperLimit = -5

    def __init__(self, ):
        self.server_name = SERVER_COMMON_NAME.replace('"', '')
        self.sort = 'none'
        self.context = {}

        self.context['radius_enable'] = False
        self.context['radius_port'] = 1812
        self.context['auth_mac'] = False
        self.context['auth_peap'] = False
        self.context['auth_adauth'] = False
        self.context['auth_cert'] = False

        self.context['clients'] = []
        self.context['client_pk'] = 0
        self.context["macs"] = []
        self.context['mac_pk'] = 0
        self.context["users"] = []
        self.context['user_pk'] = 0
        self.context["certs"] = []
        self.context['cert_pk'] = 0


    def set_sort_asc(self):
        self.sort = 'asc' if self.sort != 'asc' else 'none'


    def set_sort_dsc(self):
        self.sort = 'dsc' if self.sort != 'dsc' else 'none'


    def get_config(self):
        return self.context


    def get_clients(self):
        clients = self.context['clients']
        if self.sort == 'asc':
            clients = sorted(clients, key=lambda x: self._ip_to_int(x['name']))
        if self.sort == 'dsc':
            clients = sorted(clients, key=lambda x: self._ip_to_int(x['name']), reverse=True)
        return clients


    def get_macs(self):
        macs = self.context['macs']
        if self.sort == 'asc':
            macs = sorted(macs, key=lambda x: self._mac_to_str(x['addr']))
        if self.sort == 'dsc':
            macs = sorted(macs, key=lambda x: self._mac_to_str(x['addr']), reverse=True)
        return macs


    def get_users(self):
        users = self.context['users']
        if self.sort == 'asc':
            users = sorted(users, key=lambda x: x['name'])
        if self.sort == 'dsc':
            users = sorted(users, key=lambda x: x['name'], reverse=True)
        return users


    def get_certs(self):
        certs = self.context['certs']
        if self.sort == 'asc':
            certs = sorted(certs, key=lambda x: x['name'])
        if self.sort == 'dsc':
            certs = sorted(certs, key=lambda x: x['name'], reverse=True)
        return certs


    def _get_ip_ver(self, ip):
        try:
            socket.inet_pton(socket.AF_INET, ip)
            return 4
        except socket.error:
            try:
                socket.inet_pton(socket.AF_INET6, ip)
            except socket.error:
                return 6
        return Radius.Failed


    def _ip_to_int(self, ip):
        try:
            return int(binascii.hexlify(socket.inet_pton(socket.AF_INET, ip)), 16)
        except socket.error:
            try:
                return int(binascii.hexlify(socket.inet_pton(socket.AF_INET6, ip)), 16)
            except socket.error:
                pass
        return Radius.Failed


    def _mac_to_str(self, mac):
        str = mac.lower()
        str = str.replace(':', '')
        str = str.replace('-', '')
        return str


    def append_client(self, name, ipaddr, secret):
        clients = self.context['clients']
        if len(clients) >= RADIUS_CLIENT_MAX:
            return Radius.ErrExceededUpperLimit
        names = [e['name'] for e in clients if e['name'] == name]
        if len(names) != 0:
            return Radius.ErrExistAlready

        ipv4addr = ''
        ipv6addr = ''
        if self._get_ip_ver(ipaddr) == 4:
            ipv4addr = ipaddr
        else:
            ipv6addr = ipaddr
        clients.append({
            'pk': self.context['client_pk'],
            'name': name,
            'ipaddr': ipaddr,
            'ipv4addr': ipv4addr,
            'ipv6addr': ipv6addr,
            'secret': secret,
            'nas_type': "other",
        })
        self.context['client_pk'] += 1
        return Radius.Ok


    def update_client(self, ipaddr, secret):
        clients = self.context['clients']

        for i in range(len(clients)):
            e = clients[i]
            if e['ipaddr'] == ipaddr:
                e['secret'] = secret
                return Radius.Ok
        return Radius.ErrNotFound


    def delete_client(self, pk):
        clients = self.context['clients']
        for e in clients:
            if e['pk'] == pk:
                clients.remove(e)
                return Radius.Ok
        return Radius.ErrNotFound


    def append_mac(self, addr):
        macs = self.context['macs']
        if len(macs) >= RADIUS_MAC_MAX:
            return Radius.ErrExceededUpperLimit
        addr = addr.lower()
        if not re.match("[0-9a-f]{2}([-:]?)[0-9a-f]{2}(\\1[0-9a-f]{2}){4}$", addr):
            return Radius.ErrBadFormat
        m = addr.replace(':', '').replace('-', '')
        m = '%s:%s:%s:%s:%s:%s' % (m[0:2],m[2:4],m[4:6],m[6:8],m[8:10],m[10:12])
        names = [e['addr'] for e in macs if e['addr'] == m]
        if len(names) != 0:
            return Radius.ErrExistAlready
        macs.append({
            'pk': self.context['mac_pk'],
            'addr': m,
        })
        self.context['mac_pk'] += 1
        return Radius.Ok


    def delete_mac(self, pk):
        macs = self.context['macs']
        for e in macs:
            if e['pk'] == pk:
                macs.remove(e)
                return Radius.Ok
        return Radius.ErrNotFound


    def append_user(self, name, password, hashed):
        users = self.context['users']
        if len(users) >= RADIUS_USER_MAX:
            return Radius.ErrExceededUpperLimit
        if re.match(r'^.*[ |\t].*', name):
            return Radius.ErrBadFormat
        names = [e['name'] for e in users if e['name'] == name]
        if len(names) != 0:
            return Radius.ErrExistAlready

        users.append({
            'pk': self.context['user_pk'],
            'name': name,
            'password': password,
            'hashed': hashed,
        })
        self.context['user_pk'] += 1
        return Radius.Ok


    def update_user(self, name, password):
        users = self.context['users']
        for i in range(len(users)):
            e = users[i]
            if e['name'] == name:
                e['password'] = password
                e['hashed'] = False
                return Radius.Ok
        return Radius.ErrNotFound


    def delete_user(self, pk):
        users = self.context['users']
        for e in users:
            if e['pk'] == pk:
                users.remove(e)
                return Radius.Ok
        return Radius.ErrNotFound


    def load(self):
        self.context['clients'] = []
        self.context['client_pk'] = 0
        self.context["macs"] = []
        self.context['mac_pk'] = 0
        self.context["users"] = []
        self.context['user_pk'] = 0
        self.context["certs"] = []
        self.context['cert_pk'] = 0

        req = Request(FAL_RADIUS_PATH, filter_cmd=SUDO_CMD)
        res = req.request(Entry(ReqTypes.LOAD))

        try:
            entry_root = res.get_entrylist().get_entry().get()
        except NotFoundError:
            pass

        try:
            method = entry_root.get_node("./authenticate/method")
            self.context["radius_enable"] = entry_root.get_node("./authenticate").is_enabled()
            self.context["radius_port"] = entry_root.get_node("./authenticate/port").get_text()
            self.context["auth_mac"] = method.is_set_to_true('macaddr')
            self.context["auth_peap"] = method.is_set_to_true('peap')
            self.context["auth_adauth"] = method.is_set_to_true('adauth')
            self.context["auth_cert"] = method.is_set_to_true('eaptls')
        except:
            pass

        for authenticator in entry_root.get_node_list("./authenticate/authenticatorlist/authenticator"):
            try:
                name = authenticator.get_node("./name").get_text()
                ipaddr = authenticator.get_node("./ipaddr").get_text()
                secret = authenticator.get_node("./secretkey").get_text()
                self.append_client(name, ipaddr, secret)
            except:
                pass

        try:
            permit_root = entry_root.get_node("./permit")
        except NotFoundError:
            pass

        for macaddr in permit_root.get_node_list("./macaddrlist/macaddr"):
            try:
                addr = macaddr.get_text()
                addr = addr.lower()
                if re.match("[0-9a-f]{2}([-:]?)[0-9a-f]{2}(\\1[0-9a-f]{2}){4}$", addr):
                    m = addr.replace(':', '').replace('-', '')
                    m = '%s:%s:%s:%s:%s:%s' % (m[0:2],m[2:4],m[4:6],m[6:8],m[8:10],m[10:12])
                    self.append_mac(m)
            except:
                pass

        for user in permit_root.get_node_list("./userlist/user"):
            try:
                name = user.get_node("./name").get_text()
                password_node = user.get_node("./password")
                password = password_node.get_text()
                hashed = password_node.is_set_to_true("hashed")
                self.append_user(name, password, hashed)
            except:
                pass

        self._load_certs()


    def save(self):
        tmpl_base = \
            '<radius>\n' \
            '  <authenticate enable="%(enable)s">\n' \
            '    <port>%(port)s</port>\n' \
            '    <method macaddr="%(mac)s" peap="%(peap)s" adauth="%(adauth)s" eaptls="%(eaptls)s"/>\n'
        tmpl_client = \
            '      <authenticator>\n' \
            '        <name>%(name)s</name>\n' \
            '        <ipaddr>%(ipaddr)s</ipaddr>\n' \
            '        <secretkey>%(secret)s</secretkey>\n' \
            '      </authenticator>\n'
        tmpl_mac = \
            '      <macaddr>%(addr)s</macaddr>\n'
        tmpl_user = \
            '      <user>\n' \
            '        <name>%(name)s</name>\n' \
            '        <password hashed="%(hashed)s">%(password)s</password>\n' \
            '      </user>\n'

        xml = tmpl_base % {
            'enable':  self.context['radius_enable'],
            'port':    self.context['radius_port'],
            'mac':     self.context['auth_mac'],
            'peap':    self.context['auth_peap'],
            'adauth':  self.context['auth_adauth'],
            'eaptls':  self.context['auth_cert'],
        }

        xml += '    <authenticatorlist>\n'
        for e in self.context['clients']:
            xml += tmpl_client % {
                'name': e['name'],
                'ipaddr': e['ipaddr'],
                'secret': e['secret'],
            }
        xml += '    </authenticatorlist>\n'
        xml += '  </authenticate>\n'
        xml += '  <permit>\n'
        xml += '    <macaddrlist>\n'
        for e in self.context['macs']:
            xml += tmpl_mac % {
                'addr': e['addr'],
            }
        xml += '    </macaddrlist>\n'

        xml += '    <userlist>\n'
        for e in self.context['users']:
            xml += tmpl_user % {
                'name': e['name'],
                'password': e['password'],
                'hashed': e['hashed'],
            }
        xml += '    </userlist>\n'
        xml += '  </permit>\n'
        xml += '</radius>\n'

        req = Request(FAL_RADIUS_PATH, filter_cmd=SUDO_CMD)
        res = req.request(Entry(ReqTypes.APPLY, xml))
        result = res.get_entrylist().get_entry().get_result()
        if result != "0":
            return Radius.Failed
        return Radius.Ok


    def import_config_file(self, xml):
        req = Request(FAL_RADIUS_PATH, filter_cmd=SUDO_CMD)
        xml = re.sub('\n[ ]*', '', xml)
        res = req.request(Entry(ReqTypes.APPLY, xml))
        result = res.get_entrylist().get_entry().get_result()
        if result != "0":
            return Radius.Failed
        return Radius.Ok


    def get_path_import_csv(self):
        return RADIUS_CSV_IMPORT


    def import_csv_file(self, file):
        xml = "<radius></radius>"
        req = Request(FAL_RADIUS_PATH, filter_cmd=SUDO_CMD)
        entry = Entry(ReqTypes.IMPORT, xml)
        entry.set_file(file)
        res = req.request(entry)
        result = res.get_entrylist().get_entry().get_result()
        if result != "0":
            return Radius.Failed
        return Radius.Ok


    def get_path_export_csv(self):
        return RADIUS_CSV_EXPORT


    def export_csv_file(self, file):
        xml = "<radius></radius>"
        req = Request(FAL_RADIUS_PATH, filter_cmd=SUDO_CMD)
        entry = Entry(ReqTypes.EXPORT, xml)
        entry.set_file(file)
        res = req.request(entry)
        result = res.get_entrylist().get_entry().get_result()
        if result != "0":
            return Radius.Failed
        return Radius.Ok


    @staticmethod
    def _exec(cmd, raise_exc=True):
        ret = ""
        cmd = SUDO_CMD + " " + cmd
        proc = CmdExec(cmd.split(), stdin=NULL, stdout=PIPE, stderr=PIPE,
                       raise_exc=False, close_fds=False).execute()
        output, err = proc.communicate()
        ret = {"cmd": cmd, "returncode": proc.returncode,  "stdout": output, "stderr": err}

        if proc.wait() != 0 and raise_exc:
            raise ExecError(ret)
        return ret


    def activation(self):
        self._exec(MOUNT_CMD + " -o remount,rw /", False)
        self._exec("/sbin/insserv -dv freeradius", False)
        self._exec(MOUNT_CMD + " -o remount,ro /", False)
        self._exec("/etc/init.d/freeradius start", False)


    def invalidation(self):
        self._exec("/etc/init.d/freeradius stop", False)
        self._exec(MOUNT_CMD + " -o remount,rw /", False)
        self._exec("/sbin/insserv -r freeradius", False)
        self._exec(MOUNT_CMD + " -o remount,ro /", False)


    def append_cert(self, name, passphrase = None):
        certs = self.context['certs']
        names = [e['name'] for e in certs if e['name'] == name]
        if len(names) != 0:
            return Radius.ErrExistAlready

        if passphrase != None:
            if self._create_cert(name, passphrase) != 0:
                return Radius.Failed

        certs.append({
            'pk': self.context['cert_pk'],
            'name': name,
        })
        self.context['cert_pk'] += 1
        return Radius.Ok


    def delete_cert(self, pk):
        certs = self.context['certs']
        for e in certs:
            if e['pk'] == pk:
                 certs.remove(e)
                 self._destroy_cert(e['name'])
                 return Radius.Ok
        return Radius.ErrNotFound


    def get_path_server_cert(self):
        return RADIUS_CERT_SERVER


    def get_path_client_cert(self, pk):
        certs = self.context['certs']
        for e in certs:
            if e['pk'] == pk:
                name = e['name']
                ret = self._get_cert(name)
                if ret == Radius.Ok:
                    return RADIUS_CERT_CLIENT % {'name': name}
        return None

    def get_path_conf(self):
        return RADIUS_CONF_FILE


    def _load_certs(self):
        self.context["certs"] = []
        self.context['cert_pk'] = 0

        req = Request(FAL_RADIUS_PATH, filter_cmd=SUDO_CMD)
        xml = "<radius></radius>"
        entry = Entry(ReqTypes.LIST, xml)
        res = req.request(entry)
        result = res.get_entrylist().get_entry().get_result()
        if result != "0":
            return Radius.Failed

        root = res.get_entrylist().get_entry().get()
        certs = root.get_node_list("./radius/certificatelist/certificate")
        for c in certs:
            name = c.get_node("./name").get_text()
            if name != self.server_name:
                self.append_cert(name)
        return Radius.Ok


    def _create_cert(self, name, passphrase):
        tmpl = "<radius><certificate><name>%(name)s</name><passphrase>%(passphrase)s</passphrase></certificate></radius>"

        req = Request(FAL_RADIUS_PATH, filter_cmd=SUDO_CMD)
        xml = tmpl % {'name': name, 'passphrase': passphrase }
        entry = Entry(ReqTypes.CREATE, xml)
        file = RADIUS_CERT_CLIENT % {'name': name}
        entry.set_file(file)
        res = req.request(entry)
        result = res.get_entrylist().get_entry().get_result()
        return Radius.Ok if result == "0" else Radius.Failed


    def _destroy_cert(self, name):
        tmpl = "<radius><certificate><name>%(name)s</name></certificate></radius>"

        req = Request(FAL_RADIUS_PATH, filter_cmd=SUDO_CMD)
        xml = tmpl % {'name': name}
        entry = Entry(ReqTypes.DELETE, xml)
        res = req.request(entry)
        result = res.get_entrylist().get_entry().get_result()
        return Radius.Ok if result == "0" else Radius.Failed


    def _get_cert(self, name):
        tmpl = "<radius><certificate><name>%(name)s</name></certificate></radius>"

        req = Request(FAL_RADIUS_PATH, filter_cmd=SUDO_CMD)
        xml = tmpl % {'name': name}
        entry = Entry(ReqTypes.GET, xml)
        file = RADIUS_CERT_CLIENT % {'name': name}
        entry.set_file(file)
        res = req.request(entry)
        result = res.get_entrylist().get_entry().get_result()
        return Radius.Ok if result == "0" else Radius.Failed


if __name__ == '__main__':
    radius = Radius()
    radius.load()
    print(radius.context)
